#! /bin/python
import os
import csv
import h5py
import numpy as np
from vigra.analysis import extractRegionFeatures
from vigra.sampling import resize
from vigra.filters import distanceTransform

#
# Implementation
#


def get_common_genes(vc_genes_file_path, cells_gene_expression, med_gene_names):
    med_gene_indices = []
    vc_gene_indices = []
    common_gene_names = []
    # get the names of genes used for vc's
    with open(vc_genes_file_path) as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter=',')
        vc_gene_names = csv_reader.fieldnames
    # find a subset of genes both used for vc's and available as MEDs
    for i in range(len(vc_gene_names)):
        name = vc_gene_names[i].split('--')[0]
        if name in med_gene_names:
            med_gene_indices.append(med_gene_names.index(name))
            vc_gene_indices.append(i)
            common_gene_names.append(name)
    # from expression_by_overlap assignment extract only the subset genes
    cells_expression_subset = np.take(cells_gene_expression, med_gene_indices,
                                      axis=1)
    # from vcs_expression extract only the subset genes
    vc_expression_subset = np.loadtxt(vc_genes_file_path, delimiter=',',
                                      skiprows=1, usecols=vc_gene_indices)
    # add the null vc with no expression
    vc_expression_subset = np.insert(vc_expression_subset, 0,
                                     np.zeros(len(vc_gene_indices)),
                                     axis=0)
    return cells_expression_subset, vc_expression_subset, common_gene_names


def get_bbs(data, offset):
    shape = np.array(data.shape)
    # compute the relevant vigra region features
    # beware: for the absent labels the results are ridiculous
    features = extractRegionFeatures(data.astype('float32'), data.astype('uint32'),
                                     features=['Coord<Maximum >', 'Coord<Minimum >'])
    # compute bounding boxes from features
    mins = features['Coord<Minimum >'] - offset
    maxs = features['Coord<Maximum >'] + offset + 1
    # to prevent 'out of range' due to offsets
    mins[np.where(mins < 0)] = 0
    maxs[np.where(maxs > shape)] = shape[np.where(maxs > shape)[1]]
    # get a bb for each cell
    cell_bbs = [tuple(slice(mi, ma) for mi, ma in zip(min_, max_))
                for min_, max_ in zip(np.uint32(mins), np.uint32(maxs))]
    return cell_bbs


def get_distances(em_data, vc_data, cells_expression, vc_expression,
                  offset=10):
    num_cells = cells_expression.shape[0]
    # some labels might be lost due to downsampling
    avail_cells = np.unique(em_data)
    num_vcs = np.max(vc_data) + 1
    distance_matrix = np.full((num_cells, num_vcs), np.nan)
    bbs = get_bbs(em_data, offset)
    for cell in avail_cells:
        if cell == 0:
            continue
        bb = bbs[cell]
        cell_mask = (em_data[bb] == cell).astype("uint32")
        dist = distanceTransform(cell_mask)
        vc_roi = vc_data[bb]
        # get all the vcs within a specified distance around the cell
        vc_candidate_list = np.unique(vc_roi).astype('int')
        # select the ones not further than the offset
        vc_list = [vc for vc in vc_candidate_list
                   if np.min(dist[vc_roi == vc]) <= offset]
        cell_genes = cells_expression[cell]
        if 0 not in vc_list:
            vc_list = np.append(vc_list, 0)
        vc_genes = vc_expression[vc_list]
        # calculate the genetic distance between the cell and the vcs
        distance = np.sum(np.abs(cell_genes - vc_genes), axis=1)
        distance_matrix[cell][vc_list] = distance
    return distance_matrix


def assign_vc(distances, vc_expression):
    num_cells = distances.shape[0]
    # assign to 0 if no vcs were found at all
    assignments = [0 if np.all(np.isnan(distances[cell]))
                   else np.nanargmin(distances[cell])
                   for cell in range(num_cells)]
    cells_expr = vc_expression[assignments]
    cells_expr = np.insert(cells_expr, 0, np.arange(num_cells), axis=1)
    return cells_expr


def write_vc_assignments(segmentation_path, vc_volume_path, vc_table_path,
                         gene_table_path, table_out_path):
    em_dset = 't00000/s00/4/cells'
    cm_dset = 't00000/s00/0/cells'

    # volume file for vc's (generated from CellModels_coordinates)
    with h5py.File(vc_volume_path, 'r') as f:
        vc_data = f[cm_dset][:]

    # downsample segmentation data to the same resolution as gene data
    with h5py.File(segmentation_path, 'r') as f:
        segm_data = f[em_dset][:]
    new_shape = vc_data.shape
    downsampled_segm_data = resize(segm_data.astype('float32'), shape=new_shape,
                                   order=0).astype('uint16')

    # the table with cell expression by overlap
    with open(gene_table_path) as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter='\t')
        med_gene_names = csv_reader.fieldnames[1:]
    cells_expression = np.loadtxt(gene_table_path, delimiter='\t',
                                  skiprows=1)
    # get the genes that were both used for vcs and are in med files
    cells_expression_subset, vc_expression_subset, \
        common_gene_names = get_common_genes(vc_table_path,
                                             cells_expression[:, 1:],
                                             med_gene_names)
    # get the genetic distance from cells to surrounding vcs
    dist_matrix = get_distances(downsampled_segm_data, vc_data,
                                cells_expression_subset,
                                vc_expression_subset)
    # assign the cells to the genetically closest vcs
    cell_assign = assign_vc(dist_matrix, vc_expression_subset)
    # write down a new table
    col_names = ['label_id'] + common_gene_names
    assert cell_assign.shape[1] == len(col_names)
    with open(table_out_path, 'w') as f:
        csv_writer = csv.writer(f, delimiter='\t')
        csv_writer.writerow(col_names)
        csv_writer.writerows(cell_assign)


if __name__ == '__main__':
    segmentation_path = '/g/arendt/EM_6dpf_segmentation/EM-Prospr/em-segmented-cells-20190520-labels.h5'
    # the table generated by the first genes script
    gene_table_path = '/g/arendt/EM_6dpf_segmentation/EM-Prospr/tables/em-segmented-cells-20190520-labels_genes.csv'
    # this isn't the same file that's in the Prospr folder, I've redone the labeling, have to discuss it with Tischi
    vc_volume_path = "/g/kreshuk/zinchenk/cell_match/data/cellular-models-labels_500nm.h5"
    # this is the file from Hernando
    vc_table_path = '/g/kreshuk/zinchenk/cell_match/data/VC_to_genes.csv'
    table_out_path = os.path.splitext(gene_table_path)[0] + \
                     "_by_vc" + os.path.splitext(gene_table_path)[1]
    write_vc_assignments(segmentation_path,
                         vc_volume_path, vc_table_path,
                         gene_table_path, table_out_path)
